"""This module contains a selection of vertex signatures.
They are stored in the SIGNATURE_PROVIDERS dict under arbitrary string keys.
Each element of this should adhere to the interface of VertexSignatureProvider.
"""

import abc
import functools
import operator

import numpy as np
import scipy.spatial
import graph_tool as gt
import graph_tool.centrality
import graph_tool.topology
import graph_tool.clustering


class VertexSignatureProvider(abc.ABC):
	@property
	def eps(self):
		"""This is the signature of a nonexisting vertex."""
		return self._eps

	@property
	def metric(self):
		"""This is the function that shall be used to compute the pairwise distances
		between two signature vectors (e.g. scipy.spatial.distance.cdist)
		"""
		return self._metric
	
	@abc.abstractmethod
	def signatures(self, ltg):
		"""Computes the signature vector of the given temporal graph.

		Arguments:
		----------
		ltg : LabelledTGraph

		Returns:
		--------
		signatures : np.array of shape (t, n, k)
			Vector of vertex signatures, where t is the lifetime of the temporal graph,
			and n is the number of vertices.
			signatures[i,v,:] is the signature of vertex v in layer i
		"""
		pass

SIGNATURE_PROVIDERS = {}


class DegreesSignatureProvider(VertexSignatureProvider):
	"""Uses vertex degrees as signatures"""
	def __init__(self, **kwargs):
		self._eps = 0
		self._metric = scipy.spatial.distance.cdist

	def signatures(self, ltg):
		return degree_matrix(ltg.tgraph)[:,:,np.newaxis]

SIGNATURE_PROVIDERS["degrees"] = DegreesSignatureProvider


class Degrees2SignatureProvider(DegreesSignatureProvider):
	"""Uses vertex degrees with squared euclidean distances"""
	def __init__(self, **kwargs):
		super().__init__()
		self._metric = functools.partial(scipy.spatial.distance.cdist, metric="sqeuclidean")

SIGNATURE_PROVIDERS["degrees^2"] = Degrees2SignatureProvider


class BetweennessSignatureProvider(VertexSignatureProvider):
	"""Uses betweenness-centrality as signatures"""
	def __init__(self, **kwargs):
		self._eps = 0
		self._metric = scipy.spatial.distance.cdist

	def signatures(self, ltg):
		return betweenness_matrix(ltg.tgraph)[:,:,np.newaxis]

SIGNATURE_PROVIDERS["betweenness"] = BetweennessSignatureProvider


class DegreesBetweennessSignatureProvider(VertexSignatureProvider):
	"""Uses a combination of degrees and betweenness-centrality"""
	def __init__(self, betweenness_weight, **kwargs):
		self._betweenness_weight = betweenness_weight
		self._eps = [0, 0]
		self._metric = functools.partial(scipy.spatial.distance.cdist, metric="cityblock")

	def signatures(self, ltg):
		return np.stack([
				degree_matrix(ltg.tgraph),
				betweenness_matrix(ltg.tgraph) * self._betweenness_weight
			],
			axis=-1
		)

SIGNATURE_PROVIDERS["degrees-betweenness"] = DegreesBetweennessSignatureProvider


class DegreesUnionbetweennessSignatureProvider(VertexSignatureProvider):
	"""Uses a combination of (per-layer) degrees and the betweenness-centrality values in the union graph."""
	def __init__(self, betweenness_weight, **kwargs):
		self._betweenness_weight = betweenness_weight
		self._eps = [0, 0]
		self._metric = functools.partial(scipy.spatial.distance.cdist, metric="cityblock")

	def signatures(self, ltg):
		return np.stack(
			np.broadcast_arrays(
				degree_matrix(ltg.tgraph),
				gt.centrality.betweenness(ltg.tgraph.union_graph())[0].get_array()[np.newaxis,:] * self._betweenness_weight
			),
			axis=-1
		)

SIGNATURE_PROVIDERS["degrees-union-betweenness"] = DegreesUnionbetweennessSignatureProvider


class UnionDegreesSignatureProvider(VertexSignatureProvider):
	"""Uses the degrees of the union graph only."""
	def __init__(self, **kwargs):
		self._eps = 0
		self._metric = scipy.spatial.distance.cdist

	def signatures(self, ltg):
		tgraph = ltg.tgraph
		degrees = np.empty((tgraph.lifetime, tgraph.num_vertices()))
		vertices = tgraph.get_vertices()
		degrees[:,:] = tgraph.union_graph().get_out_degrees(vertices)
		return degrees[:,:,np.newaxis]

SIGNATURE_PROVIDERS["union-degrees"] = UnionDegreesSignatureProvider


class NeighborDegreesSignatureProvider(VertexSignatureProvider):
	"""Uses the degree, followed by the sorted list of neighbor's degrees as signature"""
	def __init__(self, max_degree, **kwargs):
		"""max_degree is the number of neighbors included in the signature"""
		self._signature_size = max_degree + 1
		self._eps = [0] * self._signature_size
		self._metric = functools.partial(scipy.spatial.distance.cdist, metric="cityblock")

	def signatures(self, ltg):
		degrees = neighbor_degrees(ltg.tgraph, self._signature_size - 1)
		delta = self._signature_size - degrees.shape[2]
		if delta > 0:
			degrees = np.pad(
				degrees,
				((0, 0), (0, 0), (0, delta)),
				mode = "constant",
				constant_values = 0
			)
		return degrees[:,:,:self._signature_size]

SIGNATURE_PROVIDERS["neighbor-degrees"] = NeighborDegreesSignatureProvider

class AvgNeighborDegreesSignatureProvider(VertexSignatureProvider):
	"""Uses the degree and the average degree of all neighbors as signature"""

	def __init__(self, **kwargs):
		self._eps = (0, 0)
		self._metric = functools.partial(scipy.spatial.distance.cdist, metric="cityblock")

	def signatures(self, ltg):
		t = ltg.tgraph.lifetime
		n = ltg.tgraph.num_vertices()
		result = np.empty((t, n, 2))
		result[:, :, 0] = degree_matrix(ltg.tgraph)
		vertices = ltg.tgraph.get_vertices()
		for i in range(t):
			graph = ltg.tgraph.layer(i)
			for v in vertices:
				avg_deg = sum(result[i, w, 0] for w in graph.get_out_neighbors(v))
				if avg_deg > 0:
					avg_deg /= result[i, v, 0]
				result[i, v, 1] = avg_deg
		return result

SIGNATURE_PROVIDERS["avg-neighbor-degree"] = AvgNeighborDegreesSignatureProvider


class ComponentsSignatureProvider(VertexSignatureProvider):
	"""Uses the size of the connected component."""

	def __init__(self, **kwargs):
		self._eps = 0
		self._metric = scipy.spatial.distance.cdist

	def signatures(self, ltg):
		return component_sizes(ltg.tgraph)[:,:,np.newaxis]

SIGNATURE_PROVIDERS["component"] = ComponentsSignatureProvider


class DegreesComponentsSignatureProvider(VertexSignatureProvider):
	"""Uses the degree and the size of the connected component."""

	def __init__(self, component_weight, **kwargs):
		self._eps = [0, 0]
		self._metric = functools.partial(scipy.spatial.distance.cdist, metric="cityblock")
		self._component_weight = component_weight

	def signatures(self, ltg):
		return np.stack((
				degree_matrix(ltg.tgraph),
				component_sizes(ltg.tgraph) * self._component_weight,
			),
			axis = -1
		)

SIGNATURE_PROVIDERS["degrees-component"] = DegreesComponentsSignatureProvider


class DegreeComponentDensitySignatureProvider(VertexSignatureProvider):
	"""Uses the degree and the density of the connected component."""

	def __init__(self, **kwargs):
		self._eps = [0, 0]
		self._metric = functools.partial(scipy.spatial.distance.cdist, metric="cityblock")

	def signatures(self, ltg):
		t = ltg.tgraph.lifetime
		result = np.zeros((t, ltg.tgraph.num_vertices(), 2))
		result[:,:,0] = degree_matrix(ltg.tgraph)
		for i in range(t):
			graph = ltg.tgraph.layer(i)
			comp_prop, comp_sizes = gt.topology.label_components(graph)
			densities = {}
			for v in graph.vertices():
				c = comp_prop[v]
				if comp_sizes[c] == 1:
					result[i, int(v), 1] = 0
				else:
					if c not in densities:
						degree_sum = np.sum(result[i, comp_prop.get_array() == c, 0])
						num_vertices = comp_sizes[c]
						densities[c] = degree_sum / (num_vertices * (num_vertices-1))
					result[i, int(v), 1] = densities[c]
		return result

SIGNATURE_PROVIDERS["degrees-density"] = DegreeComponentDensitySignatureProvider


class ClusteringCoefficientSignatureProvider(VertexSignatureProvider):
	"""Uses the local clustering coefficients."""

	def __init__(self, **kwargs):
		self._eps = [0]
		self._metric = scipy.spatial.distance.cdist

	def signatures(self, ltg):
		return clustering_coefficients(ltg.tgraph)[:,:,np.newaxis]

SIGNATURE_PROVIDERS["clustering"] = ClusteringCoefficientSignatureProvider


class NeighborhoodEdgesSignatureProvider(VertexSignatureProvider):
	"""Counts the number of edges in the closed neighborhood of each vertex."""

	def __init__(self, **kwargs):
		self._eps = [0]
		self._metric = scipy.spatial.distance.cdist

	def signatures(self, ltg):
		cluster = clustering_coefficients(ltg.tgraph)
		degrees = degree_matrix(ltg.tgraph)
		return (cluster * degrees * (degrees - 1) + degrees)[:,:,np.newaxis]

SIGNATURE_PROVIDERS["neighbor-edges"] = NeighborhoodEdgesSignatureProvider


def degree_matrix(tgraph):
	"""Given a temporal graph, returns a matrix where entry
	(i, v) contains the degree of vertex v in layer i
	"""
	degrees = np.empty((tgraph.lifetime, tgraph.num_vertices()))
	vertices = tgraph.get_vertices()
	for i in range(tgraph.lifetime):
		degrees[i, :] = tgraph.layer(i).get_out_degrees(vertices)
	return degrees


def betweenness_matrix(tgraph):
	"""Given a temporal graph, returns a matrix where entry
	(i, v) contains the betweenness of vertex v in layer i
	"""
	betweenness = np.empty((tgraph.lifetime, tgraph.num_vertices()))
	vertices = tgraph.get_vertices()
	for i in range(tgraph.lifetime):
		vp, _ = gt.centrality.betweenness(tgraph.layer(i))
		betweenness[i,:] = vp.get_array()
	return betweenness


def neighbor_degrees(tgraph, max_neighbors=float("inf")):
	n = tgraph.num_vertices()
	t = tgraph.lifetime
	k = min(max_neighbors, n-1)
	result = np.zeros((t, n, k+1))
	vertices = tgraph.get_vertices()
	for i in range(t):
		graph = tgraph.layer(i)
		degrees = graph.get_out_degrees(vertices)
		for v in vertices:
			result[i,v,0] = degrees[v]
			neighbor_degrees = sorted((degrees[u] for u in graph.get_out_neighbors(v)), reverse=True)
			result[i,v,1:1+len(neighbor_degrees)] = neighbor_degrees[:k]
	return result


def component_sizes(tgraph):
	"""Returns a np.ndarray where
	(i, v) contains the size of the connected component
	of layer i that v is part of.
	"""
	n = tgraph.num_vertices()
	t = tgraph.lifetime
	result = np.zeros((t, n))
	vertices = tgraph.get_vertices()
	for i in range(t):
		graph = tgraph.layer(i)
		comp, freq = gt.topology.label_components(graph)
		for v in vertices:
			result[i, v] = freq[comp[v]]
	return result


def clustering_coefficients(tgraph):
	"""Returns a np.ndarray in which
	(i, v) contains the local clustering coefficient 
	of vertex v in layer i.
	"""
	n = tgraph.num_vertices()
	t = tgraph.lifetime
	result = np.zeros((t, n))
	vertices = tgraph.get_vertices()
	for i in range(t):
		result[i,:] = gt.clustering.local_clustering(tgraph.layer(i)).get_array()
	return result

